/*

    Jaakko's Backscattering Simulator (JaBS)
    Copyright (C) 2021 - 2022 Jaakko Julin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    See LICENSE.txt for the full license.

 */

#ifndef CALIB_CALIBRATION_H
#define CALIB_CALIBRATION_H
#include <inttypes.h>
#include <stdlib.h>
#include <jibal_option.h>

typedef enum calibration_type {
    CALIBRATION_NONE = 0,
    CALIBRATION_ARB = 1,
    CALIBRATION_LINEAR = 2,
    CALIBRATION_POLY = 3
} calibration_type;

static const jibal_option calibration_option[] = {
        {JIBAL_OPTION_STR_NONE, CALIBRATION_NONE},
        {"arbitrary", CALIBRATION_ARB}, /* Not implemented */
        {"linear", CALIBRATION_LINEAR},
        {"polynomial", CALIBRATION_POLY},
        {NULL, 0}
};
#define CALIBRATION_PARAM_RESOLUTION (-1)
#define CALIBRATION_PARAM_OFFSET 0 /* This is zero, because offset is the first (zero'th) element in parameter arrays. Typically. */
#define CALIBRATION_PARAM_SLOPE 1

typedef struct calibration {
    calibration_type type;
    double (*f)(const void *, double);
    void *params;
    double resolution; /* Stored as FWHM in relevant SI units. Note that can be e.g. energy or time depending on detector type. */
    double resolution_variance; /* Calculated based on "resolution" before needed by detector_update(). */
} calibration;

typedef struct calibration_params_linear {
    double offset;
    double slope;
} calibration_params_linear;

typedef struct calibration_params_poly {
    size_t n;
    double *a; /* Array, size n+1, polynomial is f(x) = a[0] + a[1] * x + a[2] * x^2 + ... + a[n] * x^n */
} calibration_params_poly;

calibration *calibration_init();
void calibration_free(calibration *c);
calibration *calibration_init_linear();
calibration *calibration_init_poly(size_t n); /* n is the degree of the polynomial */
double calibration_linear(const void *params, double x);
double calibration_poly(const void *params, double x);
double calibration_none(const void *params, double x);
inline double calibration_eval(const calibration *c, double x) {return c->f(c->params, x);}
int calibration_set_param(calibration *c, int i, double value);
size_t calibration_get_number_of_params(const calibration *c);
double calibration_get_param(const calibration *c, int i); /* get i'th param in range [0..n-1], get n by  calibration_get_number_of_params()*/
double *calibration_get_param_ref(calibration *c, int i);
const char *calibration_name(const calibration *c);
char *calibration_to_string(const calibration *c);
#endif //CALIB_CALIBRATION_H
